#include <avr/interrupt.h>
#include <avr/io.h>
#include "sw_uart.h"

enum SoftwareUartStateRx {IDLE_RX, RECEIVE, RECEIVE_STOP};
enum SoftwareUartStateTx {IDLE_TX, TRANSMIT, TRANSMIT_STOP};
volatile enum SoftwareUartStateRx softwareUartStateRx = IDLE_RX;
volatile enum SoftwareUartStateTx softwareUartStateTx = IDLE_TX;
volatile uint8_t cyclesCounterRx, cyclesCounterTx; //liczba przerwa w bicie
volatile uint8_t softwareUartBitsToSend; //pozostae bity do wysania
volatile uint16_t softwareUartOutputData; //wysyany bajt
volatile uint8_t softwareUartInputData; //odbierany bajt
volatile uint8_t softwareUartNewData = 0; //flaga odebranego bajtu

void initSoftwareUart(void) {
	TCCR1B = _BV(WGM12) | _BV(CS10); //tryb CTC, brak podziau czstotliwoci
	TIMSK = _BV(OCIE1A); //przerwanie dla Output Compare
	OCR1A = F_CPU / BAUDRATE / 3 - 1; //okres licznika
	DDR_TX = _BV(PIN_TX); //pin wybrany jako TX konfigurowany jako wyjcie
	PORT_TX |= _BV(PIN_TX); //TX domylnie w stanie wysokim
}

ISR(TIMER1_COMPA_vect)
{
	static uint8_t bitsToReceive;
	static uint8_t inData;
	
	if (softwareUartStateTx == TRANSMIT) {
		if (--cyclesCounterTx == 0) {
			if(softwareUartOutputData & 1) PORT_TX |= _BV(PIN_TX); else PORT_TX &= ~_BV(PIN_TX); //data bit
			softwareUartOutputData >>= 1;
			cyclesCounterTx = 3;
			if (--softwareUartBitsToSend == 0) {
				softwareUartStateTx = TRANSMIT_STOP;
			}
		}
	}
	
	if (softwareUartStateTx == TRANSMIT_STOP) {
		if (--cyclesCounterTx == 0) {
			softwareUartStateTx = IDLE_TX;
		}
	}
	
	if ((softwareUartStateRx == IDLE_RX) && !(PORT_RX & _BV(PIN_RX))) {
		cyclesCounterRx = 4;
		softwareUartStateRx = RECEIVE;
		bitsToReceive = 8;
		inData = 0;
	}
	
	if (softwareUartStateRx == RECEIVE) {
		if (--cyclesCounterRx == 0) {
			inData >>= 1;
			if (PORT_RX & _BV(PIN_RX)) inData |= 0x80;
			cyclesCounterRx = 3;
			if (--bitsToReceive == 0) {
				softwareUartStateRx = RECEIVE_STOP;
				softwareUartInputData = inData;
				softwareUartNewData = 1;
			}
		}
	}
	
	if (softwareUartStateRx == RECEIVE_STOP) {
		if (--cyclesCounterRx == 0) {
			softwareUartStateRx = IDLE_RX;
		}
	}
}

void softwareUartSendByte(uint8_t data) {
	while(softwareUartStateTx != IDLE_TX);
	cyclesCounterTx = 1;
	softwareUartBitsToSend = 10;
	softwareUartOutputData = ((uint16_t)data << 1) | 0b1000000000;
	softwareUartStateTx = TRANSMIT;
}

void softwareUartSendBytes(uint8_t data[], uint16_t size) {
	for (uint16_t i = 0; i < size; i++) {
		softwareUartSendByte(data[i]);
	}
}

void softwareUartSendString(char str[]) {
	uint16_t i = 0;
	while (str[i]) {
		softwareUartSendByte(str[i++]);
	}
}